defmodule CodeLock do

  @moduledoc """
  Implementation of Erlang code_lock FSM in Elixir.

  """

  @name :code_lock

  def start(code \\ 'abc') do
    start_link(code)
  end

  def start_link(code) do
    :gen_statem.start_link({:local, @name}, __MODULE__, code, [])
  end

  def stop() do
    :gen_statem.stop(@name)
  end

  def button(button) do
    :gen_statem.cast(@name, {:button, button})
  end

  @spec init(charlist()) :: {:ok, atom(), Map.t()}
  def init(code) do
    do_lock()
    data = %{:code => code, :length => length(code), :buttons => []}
    {:ok, :locked, data}
  end

  def callback_mode() do
    :state_functions
  end

  def locked(:cast, {:button, button}, data) do
    %{:code => code, :length => codelength, :buttons => buttons} = data
    newbuttons = if length(buttons) < codelength do
      buttons
    else
      tl(buttons)
    end
    newbuttons1 = newbuttons ++ button
    if newbuttons1 == code do
      IO.puts("unlocking")
      do_unlock()
      data1 = Map.put(data, :buttons, [])
      {:next_state, :open, data1, [{:state_timeout, 10000, :lock}]}
    else
      IO.puts("staying locked")
      data1 = Map.put(data, :buttons, newbuttons1)
      {:next_state, :locked, data1}
    end
  end

  def open(:state_timeout, :lock, data) do
    IO.puts("(open)")
    do_lock()
    {:next_state, :locked, data}
  end
  def open(:cast, {:button,_}, data) do
    {:next_state, :open, data}
  end

  def do_lock() do
    IO.puts("Lock")
  end

  def do_unlock() do
    IO.puts("Unlock")
  end

  def terminate(_reason, state, _data) do
    if state == :locked do
      do_lock()
    end
    :ok
  end

end
